/**********************************************************************
  Filename    : Camera Remote Bluetooth IoT
  Description : Control android app Camera Remote Bluetooth remotely.
                Change camera modes: photo or video.
                Change flash modes: torch or auto
                Record video.
                Take photo.
  Version     : ESP32 CAM AI Thinker and for TFT ST7789 240x240 Color Display            
  Auther      : www.busywww.com   busyweb@gmail.com
  Modification: 2021/10/30
**********************************************************************/
#include "soc/soc.h"
#include "soc/rtc_cntl_reg.h"

#include "BluetoothSerial.h"
#include "string.h"

#include "esp_timer.h"
#include "img_converters.h"
#include "fb_gfx.h"
#include "driver/ledc.h"
#include "sdkconfig.h"
#include "esp_camera.h"

#include <Arduino.h>
#include <SPI.h>
#include <Adafruit_GFX.h>
//#include <Adafruit_ST7789.h>
#include "Arduino_ST7789_Comp.h" // Hardware-specific library for ST7789
#include <TJpg_Decoder.h>

#include "FS.h"
#include "SD_MMC.h"


#if defined(ARDUINO_ARCH_ESP32) && defined(CONFIG_ARDUHAL_ESP_LOG)
#include "esp32-hal-log.h"
#define TAG ""
#else
#include "esp_log.h"
static const char *TAG = "camera_httpd";
#endif

#ifdef CONFIG_LED_ILLUMINATOR_ENABLED
int led_duty = 0;
bool isStreaming = false;
#ifdef CONFIG_LED_LEDC_LOW_SPEED_MODE
#define CONFIG_LED_LEDC_SPEED_MODE LEDC_LOW_SPEED_MODE
#else
#define CONFIG_LED_LEDC_SPEED_MODE LEDC_HIGH_SPEED_MODE
#endif
#endif

#define CAMERA_MODEL_AI_THINKER // Has PSRAM
#include "camera_pins.h"


typedef struct
{
    size_t size;  //number of values used for filtering
    size_t index; //current value index
    size_t count; //value count
    int sum;
    int *values; //array to be filled with values
} ra_filter_t;


#ifdef CONFIG_LED_ILLUMINATOR_ENABLED
void enable_led(bool en)
{ // Turn LED On or Off
    int duty = en ? led_duty : 0;
    if (en && isStreaming && (led_duty > CONFIG_LED_MAX_INTENSITY))
    {
        duty = CONFIG_LED_MAX_INTENSITY;
    }
    ledc_set_duty(CONFIG_LED_LEDC_SPEED_MODE, CONFIG_LED_LEDC_CHANNEL, duty);
    ledc_update_duty(CONFIG_LED_LEDC_SPEED_MODE, CONFIG_LED_LEDC_CHANNEL);
    ESP_LOGI(TAG, "Set LED intensity to %d", duty);
}
#endif


#define MCU_MODE 3    //1: ESP32 WROOM 32, 2: ESP32 WROVER, 3: ESP32 CAM
#define DISPLAY_COLOR_MODE 2      // 0: no display, 1: Monochrome (SD1306, black and white), 2: Color (ST7789)
#define SCREEN_WIDTH  240 
#define SCREEN_HEIGHT 240 

// ESP32 CAM: SPI Connection *****************************************
// Ueer selective pins: DC, RESET
// ST7789 ***************************************************************

#define TFT_CS         -1     //
#define TFT_RST        2      //1      //9 // Or set to -1 and connect to Arduino RESET pin
#define TFT_DC         13     //13
#define TFT_MOSI       14     //3     //11  // Data out
#define TFT_SCLK       15     //4     //13  // Clock out
//// Initialize Adafruit ST7789 TFT library
Arduino_ST7789 tft = Arduino_ST7789(TFT_DC, TFT_RST, TFT_MOSI, TFT_SCLK);

// Hardware  SPI
//#define TFT_CS         -1      // set pin number for library calling, but do not connect to mcu for ST7789 display and SPI mode (may need to connet for other display) 
//#define TFT_RST        4      // any available pin
//#define TFT_DC         15     // any available pin
//#define TFT_MOSI       23     // Data out
//#define TFT_SCLK       18 
// For 1.14", 1.3", 1.54", 1.69", and 2.0" TFT with ST7789:
// Hardware SPI
//Adafruit_ST7789 tft = Adafruit_ST7789(TFT_CS, TFT_DC, TFT_RST);

// ***********************************************************************


// Bluetooth Status
int CONNECTED = 0;
BluetoothSerial SerialBT;

// variables for application
boolean DEBUG = false;
bool imageDataProcessing = false;
String dataValues[15]; 
boolean needDisplayRefresh = false; 
bool holdSendingFrame = false;

#define PIN_LED_FLASH     4    // find one available gpio for flash, if no hardware flash is available.
#define PIN_BUTTON_PHOTO  12
#define PIN_BUTTON_FLASH  1
int ACTION_STATUS = 0;
int FLASH_MODE = 0;

String filelist;
long file_number = 0;
long file_count = 0;
uint64_t cardSize = 0;
uint64_t totalBytes = 0;       // = SPIFFS.totalBytes();
uint64_t usedBytes = 0;        // = SPIFFS.usedBytes();
uint64_t availableBytes = 0;   // totalBytes - usedBytes;


// This next function will be called during decoding of the jpeg file to
// render each block to the TFT.  If you use a different TFT library
// you will need to adapt this function to suit.
bool tft_output(int16_t x, int16_t y, uint16_t w, uint16_t h, uint16_t* bitmap)
{
   // Stop further decoding as image is running off bottom of screen
  if ( y >= tft.height() ) return 0;

  //tft.drawRGBBitmap(x, y, bitmap, w, h);
 
  //tft.drawImageF(x, y, w, h, bitmap);
  tft.drawImage(x, y, w, h, bitmap);
  
  // Return 1 to decode next block
  return 1;
}

// SPIFFS related functions ****************************************

void SPIFFS_SpaceInformation() {
  totalBytes = SPIFFS.totalBytes();
  usedBytes = SPIFFS.usedBytes();
  availableBytes = totalBytes - usedBytes;
  //Serial.println("Total: " + String(totalBytes) + ", Used: " + String(usedBytes) + ", Available: " + String(availableBytes));
  Serial.printf("Total space: %lluKB\n", totalBytes / (1024));
  Serial.printf("Used space: %lluKB\n", usedBytes / (1024));
  Serial.printf("Available space: %lluKB\n", (totalBytes - usedBytes) / (1024));  
}

void filesList() {
  filelist = "";
  file_count = 0;
  String fileNum = "0";
  int fileNumInt = 0;

  boolean needAdjustForNumber = false;
  if (file_number == 0) {
    needAdjustForNumber = true;
  }

  fs::File root = SPIFFS.open("/");
  File file = root.openNextFile();
  while (file) {
    String fileName = file.name();
    //Serial.println(fileName);
    if (fileName.startsWith("image")) {
      file_count++;
      filelist = filelist + ":" + file.name();
      fileNum = fileName;
      fileNum.replace("image_", "");
      fileNum.replace(".jpg", "");
      fileNumInt = fileNum.toInt();
      if (fileNumInt > file_number) {
        file_number = fileNumInt;
      }
    }
    file = root.openNextFile();
  }
  if (file) {
    file.close();
  }
  if (root) {
    root.close();
  }
  Serial.println(filelist);
  SPIFFS_SpaceInformation();
}

void sendFileList() {
  holdSendingFrame = true;
  filelist = "";
  file_count = 0;

  filesList();

  SerialBT.print("filelist");
  if (file_count > 0) {
    SerialBT.print(filelist);
  } else {
    SerialBT.print(":");
  }
   // send new line
  SerialBT.write(13);
  SerialBT.write(10); 

  delay(1000);    

  holdSendingFrame = false;
}

void sendFileContents(String dataString, boolean forList) {
  holdSendingFrame = true;

  parseData(dataString);
  String fileName = dataValues[1];
  
  if (forList) {
    fileName.replace("image", "thumb");
  }
   
  File file = SPIFFS.open("/" + fileName);
  long fileSize = 0;
  if (file) {
    fileSize = file.size();

    Serial.println("Sending file data: " + fileName + ", " + String(fileSize));
    if (forList) {
      SerialBT.print("imagecontents:"); 
      //fileName.replace("thumb", "image");
      SerialBT.print(fileName + ":" + String(fileSize) + ":");
    } else {
      SerialBT.print("filecontents:"); 
      SerialBT.print(fileName + ":" + String(fileSize) + ":");
    }
     // send new line
    SerialBT.write(13);
    SerialBT.write(10); 
  
    // send contents
    while(file.available()) {
      SerialBT.write(file.read());
    }
     // send new line
    SerialBT.write(13);
    SerialBT.write(10);    
    
  } else {
    fileSize = 0;
    Serial.println("Sending file data: " + fileName + ", " + String(fileSize));    
  }

  if (file) {
    file.close();
  }
  delay(500);

  holdSendingFrame = false;
}

void deleteFileFromRemote(String dataString) {
  holdSendingFrame = true;
  parseData(dataString);
  String fileName = dataValues[1]; 
  SPIFFS.remove("/" + fileName);
  fileName.replace("image", "thumb");
  SPIFFS.remove("/" + fileName);

  // send processing result
  SerialBT.print("filedeleteresult:" + fileName + ":success:");  

   // send new line
  SerialBT.write(13);
  SerialBT.write(10);
  
  holdSendingFrame = false;

  filesList();

  tft.fillRect(0, 0, 240, 240, BLACK);
  writeText(40, 60, 1, WHITE, "Camera Remote Bluetooth");
  writeText(40, 100, 2, WHITE, "Connected.");
  writeText(40, 160, 2, WHITE, String(file_count) + " images");
}

void getLastFileNumber() {
  //file_number = millis();
  String fileNum = "0";
  File file = SPIFFS.open("/lastnum.txt");
  if (!file) {
    file = SPIFFS.open("/lastnum.txt", FILE_WRITE);
    file.println("0");
    file.close();
  } else {
    while(file.available()) {
      fileNum = file.readStringUntil('\n');
    }
    file_number = fileNum.toInt();
    file.close();
  }
  Serial.println("Last file number: " + String(file_number));
}
void saveLastFileNumber() {
  SPIFFS.remove("/lastnum.txt");
  File file = SPIFFS.open("/lastnum.txt", FILE_WRITE);
  file.println(String(file_number));
  file.close();
  Serial.println("Last file number saved. " + String(file_number));
}

void getFileNumberForImage(String dataString) {
  // dataString: frameget:xxxxx: from remote command
  parseData(dataString);
  long fileNum = dataValues[1].toInt();   //atol(dataValues[1]);  

  if (fileNum == 0) {
    file_number = 0;
    saveLastFileNumber();
  } else {
    if (file_number <= fileNum) {
      file_number = fileNum;
    }    
  }   
}
// ****************************************************************************

void setup() {

  pinMode(PIN_BUTTON_PHOTO, INPUT_PULLUP);
  pinMode(PIN_BUTTON_FLASH, INPUT_PULLUP);
  
  Serial.begin(115200);

//  tft.init(240, 240, SPI_MODE3);           // Init ST7789 240x240, SPI_MODE3
  tft.init(240, 240);           
  tft.fillScreen(BLACK); 
  delay(100);
  TJpgDec.setJpgScale(1);
  // The decoder must be given the exact name of the rendering function above
  TJpgDec.setCallback(tft_output);
  delay(100);
  writeText(40, 100, 2, WHITE, "Camera Remote");
  writeText(40, 130, 2, WHITE, "Bluetooth");
  delay(2500);


  WRITE_PERI_REG(RTC_CNTL_BROWN_OUT_REG, 0);    //disable brownout detector   ds

//  SerialBT.begin("Camera Remote Bluetooth");    //Bluetooth device name
  SerialBT.begin("Bluetooth Camera", false);    //Bluetooth device name, slave mode (connect from android device) 
  SerialBT.register_callback(callback);
  
  Serial.println("\nThe camera started, now you can connect (paired) from android remote mode by bluetooth!");
  
  delay(1000);
  WRITE_PERI_REG(RTC_CNTL_BROWN_OUT_REG, 1);    //enable brownout detector

  if (CONNECTED) {
    // no codes will be run here
  }
  
  camera_config_t config;
  config.ledc_channel = LEDC_CHANNEL_0;
  config.ledc_timer = LEDC_TIMER_0;
  config.pin_d0 = Y2_GPIO_NUM;
  config.pin_d1 = Y3_GPIO_NUM;
  config.pin_d2 = Y4_GPIO_NUM;
  config.pin_d3 = Y5_GPIO_NUM;
  config.pin_d4 = Y6_GPIO_NUM;
  config.pin_d5 = Y7_GPIO_NUM;
  config.pin_d6 = Y8_GPIO_NUM;
  config.pin_d7 = Y9_GPIO_NUM;
  config.pin_xclk = XCLK_GPIO_NUM;
  config.pin_pclk = PCLK_GPIO_NUM;
  config.pin_vsync = VSYNC_GPIO_NUM;
  config.pin_href = HREF_GPIO_NUM;
  config.pin_sscb_sda = SIOD_GPIO_NUM;
  config.pin_sscb_scl = SIOC_GPIO_NUM;
  config.pin_pwdn = PWDN_GPIO_NUM;
  config.pin_reset = RESET_GPIO_NUM;
  config.xclk_freq_hz = 20000000;
  config.pixel_format = PIXFORMAT_JPEG;
  
  // if PSRAM IC present, init with UXGA resolution and higher JPEG quality
  //                      for larger pre-allocated frame buffer.
  if(psramFound()){
    config.frame_size = FRAMESIZE_UXGA;
    config.jpeg_quality = 10;
    config.fb_count = 2;
  } else {
    config.frame_size = FRAMESIZE_SVGA;
    config.jpeg_quality = 12;
    config.fb_count = 1;
  }

#if defined(CAMERA_MODEL_ESP_EYE)
  pinMode(13, INPUT_PULLUP);
  pinMode(14, INPUT_PULLUP);
#endif
#if defined(CAMERA_MODEL_AI_THINKER)
//  pinMode(2, OUTPUT);
//  pinMode(15, OUTPUT);
//  pinMode(13, OUTPUT);
//  pinMode(14, OUTPUT);
  pinMode(PIN_LED_FLASH, OUTPUT);
#endif

  // camera init
  esp_err_t err = esp_camera_init(&config);
  if (err != ESP_OK) {
    Serial.printf("Camera init failed with error 0x%x", err);
    return;
  } else {
    Serial.println("Camera initialized successfully.");
  }

  sensor_t * s = esp_camera_sensor_get();
  // initial sensors are flipped vertically and colors are a bit saturated
  if (s->id.PID == OV2640_PID) {
    //s->set_vflip(s, 1); // flip it back
    s->set_brightness(s, 1); // up the brightness just a bit
    //s->set_saturation(s, 1); // bit of the saturation
  }
  // drop down frame size for higher initial frame rate
  s->set_framesize(s, FRAMESIZE_HQVGA);

#if defined(CAMERA_MODEL_M5STACK_WIDE) || defined(CAMERA_MODEL_M5STACK_ESP32CAM)
  s->set_vflip(s, 1);
  s->set_hmirror(s, 1);
#endif

// SPIFFS
  if (!SPIFFS.begin()) {
    Serial.println("SPIFFS initialisation failed!");
    SPIFFS.begin(true);// Formats SPIFFS - could lose data https://github.com/espressif/arduino-esp32/issues/638
  } else {
    Serial.println("SPIFFS initialized.");
    delay(200);
    getLastFileNumber();
    filesList();  
  }

}

void changeFrameSize(String dataString) {

  parseData(dataString);

  String frameSize = dataValues[1];
  framesize_t fs = FRAMESIZE_HQVGA;
  
  if (frameSize.equalsIgnoreCase("small")) {
    fs = FRAMESIZE_QQVGA;  //QQVGA(160x120), FRAMESIZE_QQVGA;
  } else if (frameSize.equalsIgnoreCase("medium")) {
    fs = FRAMESIZE_HQVGA;  //HQVGA(240x176), FRAMESIZE_QVGA;
  } else if (frameSize.equalsIgnoreCase("large")) {
    fs = FRAMESIZE_QVGA;  //QVGA(320x240), FRAMESIZE_HVGA;  //FRAMESIZE_VGA;
  }

  sensor_t *s = esp_camera_sensor_get();
  int res = 0;
  res = s->set_framesize(s, fs);

}

void callback(esp_spp_cb_event_t event, esp_spp_cb_param_t *param) {
  if(event == ESP_SPP_SRV_OPEN_EVT){
    Serial.println("Client Connected");
    //digitalWrite(LED,HIGH);
    CONNECTED = true;

    // clear display
    needDisplayRefresh = true;
    delay(500);
    tft.fillScreen(BLACK); 
    delay(500);
    writeText(40, 60, 1, WHITE, "Camera Remote Bluetooth");
    writeText(40, 100, 2, WHITE, "Connected.");
    writeText(40, 160, 2, WHITE, String(file_count) + " images");
  }
 
  if(event == ESP_SPP_CLOSE_EVT ){
    Serial.println("Client disconnected");
    //digitalWrite(LED,LOW);
    CONNECTED = false;

    // resize camera frame size
    delay(500);
    changeFrameSize("framesize:medium");
    delay(1000);
    needDisplayRefresh = true;
  }
}

// display text
void writeText(uint16_t x, uint16_t y, uint16_t textSize, int textColor, String textString) {
  tft.setTextWrap(true);
  tft.setCursor(x, y);
  tft.setTextSize(textSize);
  tft.setTextColor(textColor);
  tft.println(textString);
}

// clears the data value array
void clearDataValues() {
  int numberOfValues = 10;    //(int)sizeof(dataValues)/sizeof("hello");
  for (int i = 0; i < numberOfValues; i++) {
    dataValues[i] = "";
  }
}

void parseData(String dataString) {
  clearDataValues();

  // data delimeter: ":"
  char delimeter = ':';
  int findStart = 0;
  int findIndex = 0;
  int findCount = 0;

  findIndex = dataString.indexOf(delimeter, findStart);
  if (findIndex > 0) {
    dataValues[findCount] = dataString.substring(findStart, findIndex);
    findCount ++;
    findStart = findIndex + 1;

    while (findIndex > 0) {
      findIndex = dataString.indexOf(delimeter, findStart);
      if (findIndex > 0) {
        dataValues[findCount] = dataString.substring(findStart, findIndex);
        findCount ++;
        findStart = findIndex + 1;        
      }
    } 
  }
}

void changeFlashMode(String dataString) {
  parseData(dataString);

  String flashMode = dataValues[1];
  if (flashMode.equalsIgnoreCase("on")) {
    FLASH_MODE = 1;
#ifdef CONFIG_LED_ILLUMINATOR_ENABLED
    enable_led(true);
#endif
#if defined(CAMERA_MODEL_AI_THINKER)
    digitalWrite(PIN_LED_FLASH, HIGH);
#endif
  } else if (flashMode.equalsIgnoreCase("off")) {
    FLASH_MODE = 0;
#ifdef CONFIG_LED_ILLUMINATOR_ENABLED
    enable_led(false);
#endif 
#if defined(CAMERA_MODEL_AI_THINKER)
    digitalWrite(PIN_LED_FLASH, LOW);
#endif   
  }
}

void changeExposure(String dataString) {
  Serial.println(dataString);
  parseData(dataString);

  String exposure = dataValues[1];
  sensor_t *s = esp_camera_sensor_get();
  int res = 0;
  res = s->set_exposure_ctrl(s, 1);   // 0: disable, 1: enable
  //res = s->set_brightness(s, exposure.toInt());   // -2 to 2
  res = s->set_ae_level(s, exposure.toInt());     // -2 to 2
}

void changeEffect(String dataString) {
  Serial.println(dataString);
  parseData(dataString);
  String effect = dataValues[1];
  sensor_t *s = esp_camera_sensor_get();
  int res = 0;
  res = s->set_special_effect(s, effect.toInt());     // (0 - No Effect, 1 - Negative, 2 - Grayscale, 3 - Red Tint, 4 - Green Tint, 5 - Blue Tint, 6 - Sepia)
}

void processAction() {

  int photo = digitalRead(PIN_BUTTON_PHOTO);
  int flash = digitalRead(PIN_BUTTON_FLASH);

  if (DEBUG) {
    Serial.println("photo: " + String(photo) + " flash: " + String(flash) + " FLASH_MODE: " + String(FLASH_MODE));
  }
  
  if (ACTION_STATUS == 0) {

    if (!photo) {
      // take frame and save
      if (!CONNECTED) {
        takePictureAndSave();
      } else {
      }     
    } else if (!flash) {
      // flash
      if (FLASH_MODE == 1) {
        //changeFlashMode("flash:off");
        digitalWrite(PIN_LED_FLASH, LOW);
        FLASH_MODE = 0;
      } else {
        //changeFlashMode("flash:on");
        digitalWrite(PIN_LED_FLASH, HIGH);
        FLASH_MODE = 1;
      }
    }   

     ACTION_STATUS = 1;
     delay(500);
  }

  if (ACTION_STATUS == 1) {
     if (photo && flash) {
      ACTION_STATUS = 0;
      delay(10);
     }
  }
    
}

// when not connected,
// display camera frame to the display
long refreshDisplayTime = 0;
void displayCamImage(void)
{
    size_t _jpg_buf_len = 0;
    uint8_t *_jpg_buf = NULL;    
    camera_fb_t *fb = esp_camera_fb_get();

    if (!fb) {
        //ESP_LOGE(TAG, "Frame buffer could not be acquired");
        Serial.println("Frame buffer could not be acquired");
        delay(200);
        return;
    }

    _jpg_buf_len = fb->len;
    _jpg_buf = fb->buf;    
    uint16_t w = 0, h = 0;
    TJpgDec.getJpgSize(&w, &h, fb->buf, fb->len);
    if (DEBUG) {
      Serial.print("- Width = "); Serial.print(fb->width); Serial.print(", height = "); Serial.println(fb->height);         
    }
    TJpgDec.drawJpg(0, 0, _jpg_buf, _jpg_buf_len * sizeof(byte));

    esp_camera_fb_return(fb);

    if (needDisplayRefresh) {
      tft.fillRect(0, 0, 240, 240, BLACK);
      refreshDisplayTime = millis();
      needDisplayRefresh = false;
    } else {
      if (millis() - refreshDisplayTime < 2000) {
        tft.fillRect(0, 0, 240, 240, BLACK);
      } else {
        refreshDisplayTime = 0;
      }
    }
    writeText(40, 190, 2, WHITE, "Not connected.");
    writeText(40, 210, 2, WHITE, String(file_count) + " images");
}

// send camera frame data to the remote android app
// when the bluetooth is connected
esp_err_t camera_capture(){
    camera_fb_t * fb = NULL;
    size_t _jpg_buf_len = 0;
    uint8_t *_jpg_buf = NULL;
  
    //capture a frame
    fb = esp_camera_fb_get();
    if (!fb) {
        //ESP_LOGE(TAG, "Frame buffer could not be acquired");
        Serial.println("Frame buffer could not be acquired");
        delay(200);
        return ESP_FAIL;
    }

    _jpg_buf_len = fb->len;
    _jpg_buf = fb->buf;

    if (fb)
    {
        if (DEBUG) {
          Serial.print("width: ");
          Serial.print(fb->width);
          Serial.print(", height: ");
          Serial.print(fb->height);
          Serial.print(", len: ");
          Serial.println(_jpg_buf_len);          
        } 

        if (CONNECTED) {
          
          // send frame data header
          SerialBT.print("frame_data");
          SerialBT.print(":width:");
          SerialBT.print(fb->width);
          SerialBT.print(":height:");
          SerialBT.print(fb->height);
          SerialBT.print(":len:");
          SerialBT.println(_jpg_buf_len);

          // send frame data
          //frameData = frameData[_jpg_buf_len];
          //memcpy(frameData, _jpg_buf, _jpg_buf_len);  
          
          //frameData = (uint8_t*)malloc(_jpg_buf_len * sizeof(uint8_t));
          //memmove(frameData, _jpg_buf, _jpg_buf_len);  
          
          //delay(10);       
          //SerialBT.write(frameData, _jpg_buf_len);         
          SerialBT.write(_jpg_buf, _jpg_buf_len);         

          // send new line
          SerialBT.write(13);
          SerialBT.write(10);
        }

        esp_camera_fb_return(fb);
        fb = NULL;
        _jpg_buf = NULL;

        
    } else if (_jpg_buf)
    {
        free(_jpg_buf);
        _jpg_buf = NULL;
    }    

    delay(50);
    return ESP_OK;
}

// remote take photo command processing
// get largest camera frame and send to remote android app
// android app saves photo in the storage
void takePictureAndSend() {
  holdSendingFrame = true;
  sensor_t *s = esp_camera_sensor_get();
  //framesize_t fs = s->status.framesize;
  int res = 0;
  res = s->set_framesize(s, FRAMESIZE_UXGA);  //SVGA800x600, change accordingly

//  delay(500);

  camera_fb_t * fb = NULL;
  size_t _jpg_buf_len = 0;
  uint8_t *_jpg_buf = NULL;

  // wait for the auto exposure sets
  for (int i = 0; i < 9; i++) {
    fb = esp_camera_fb_get();
    esp_camera_fb_return(fb);
    if (fb) {
      fb = NULL;
      _jpg_buf = NULL;
    } else if (_jpg_buf) {
      free(_jpg_buf);
      _jpg_buf = NULL;      
    }
    delay(100);  
  }

  
  fb = esp_camera_fb_get();
  if (!fb) {
    //ESP_LOGE(TAG, "Frame buffer could not be acquired");
    Serial.println("Frame buffer could not be acquired");
      
    //send failed status message
    // send frame data header
    SerialBT.print("picture_data");
    SerialBT.print(":width:0");
    SerialBT.print(":height:0");
    SerialBT.print(":len:0:");
     // send new line
    SerialBT.write(13);
    SerialBT.write(10);  
     
  } else {

    _jpg_buf_len = fb->len;
    _jpg_buf = fb->buf;    
    
    // send frame data header
    SerialBT.print("picture_data");
    SerialBT.print(":width:");
    SerialBT.print(fb->width);
    SerialBT.print(":height:");
    SerialBT.print(fb->height);
    SerialBT.print(":len:");
    SerialBT.println(_jpg_buf_len);
         
    SerialBT.write(_jpg_buf, _jpg_buf_len);   
    // send new line
    SerialBT.write(13);
    SerialBT.write(10);  

    esp_camera_fb_return(fb);
    if (fb) {
      fb = NULL;
      _jpg_buf = NULL;
    } else if (_jpg_buf) {
      free(_jpg_buf);
      _jpg_buf = NULL;      
    }
  }

  sensor_t * s3 = esp_camera_sensor_get();
  s3->set_framesize(s3, FRAMESIZE_HQVGA);  
    
  delay(1000);
  
  holdSendingFrame = false;
}

// local camera button (take photo) action processing
// take camera frame (size of XGA(1024x768))
// and saves the photo into the board flash memory SPIFFS
// about 20 photos could be saved in the flash memory
// Remote android app could retrieve the photos and saves into the android storage.
// Remote android app could delete ESP32 Cam stored photos on demand.
void takePictureAndSave() {
  holdSendingFrame = true;

  getLastFileNumber();
  filesList();

  if (availableBytes < (250 * 1024)) {
    holdSendingFrame = true;    
    
    if (!CONNECTED) {
      needDisplayRefresh = true;
      tft.fillRect(0, 0, 240, 240, BLACK);
      writeText(40, 190, 2, WHITE, "Space is full.");
      writeText(40, 210, 2, WHITE, String(file_count) + " images");
      delay(2000);
      tft.fillRect(0, 0, 240, 240, BLACK);
      writeText(40, 190, 2, WHITE, "Not connected.");
      writeText(40, 210, 2, WHITE, String(file_count) + " images");
    } else {
      tft.fillRect(0, 0, 240, 240, BLACK);
      writeText(40, 100, 2, WHITE, "Space is full.");
      writeText(40, 160, 2, WHITE, String(file_count) + " images");
      delay(2000);
      tft.fillRect(0, 0, 240, 240, BLACK);
      writeText(40, 60, 1, WHITE, "Camera Remote Bluetooth");
      writeText(40, 100, 2, WHITE, "Connected.");
      writeText(40, 160, 2, WHITE, String(file_count) + " images");     
    }
    
    delay(500);
    takePictureAndSend();
    
    return;
  }
  file_number += 1;

  camera_fb_t * fb = NULL;
  size_t _jpg_buf_len = 0;
  uint8_t *_jpg_buf = NULL;

  //save thumb for remote file listing
  sensor_t *s = esp_camera_sensor_get();
  //framesize_t fs = s->status.framesize;
  int res = 0;
  res = s->set_framesize(s, FRAMESIZE_QQVGA);  //VGA, UXGA
  delay(100);

  fb = esp_camera_fb_get();
  if (!fb) {
    //ESP_LOGE(TAG, "Frame buffer could not be acquired");
    Serial.println("Frame buffer could not be acquired");    
  } else {

    _jpg_buf_len = fb->len;
    _jpg_buf = fb->buf;    

    // thumb file, save to Flash Memory
    String full_filename = "/thumb_" + String(file_number) + ".jpg";
    File file = SPIFFS.open(full_filename.c_str(), FILE_WRITE);
    if (file) {
      file.write(fb->buf, fb->len); // payload (image), payload length
      file.close();
    } else {
      Serial.println("Failed to create file.");
    }
  }
  esp_camera_fb_return(fb);
  if (fb) {
    fb = NULL;
    _jpg_buf = NULL;
  } else if (_jpg_buf) {
    free(_jpg_buf);
    _jpg_buf = NULL;      
  }

  // large size image 
  // large size image 
  sensor_t *s2 = esp_camera_sensor_get();
  res = 0;
  res = s2->set_framesize(s2, FRAMESIZE_SVGA);  //800x600, space limitation: change accordingly

  // wait for the auto exposure sets
  for (int i = 0; i < 9; i++) {
    fb = esp_camera_fb_get();
    esp_camera_fb_return(fb);
    if (fb) {
      fb = NULL;
      _jpg_buf = NULL;
    } else if (_jpg_buf) {
      free(_jpg_buf);
      _jpg_buf = NULL;      
    }
    delay(100);  
  } 
  
  fb = esp_camera_fb_get();
  if (!fb) {
    //ESP_LOGE(TAG, "Frame buffer could not be acquired");
    Serial.println("Frame buffer could not be acquired");
  } else {

    _jpg_buf_len = fb->len;
    _jpg_buf = fb->buf;    

    // save to Flash Memory
    String full_filename = "/image_" + String(file_number) + ".jpg";
    File file = SPIFFS.open(full_filename.c_str(), FILE_WRITE);
    if (file) {
      file.write(fb->buf, fb->len); // payload (image), payload length
      file.close();
    } else {
      Serial.println("Failed to create file.");
    }

    esp_camera_fb_return(fb);
    if (fb) {
      fb = NULL;
      _jpg_buf = NULL;
    } else if (_jpg_buf) {
      free(_jpg_buf);
      _jpg_buf = NULL;      
    }
   
  }

  delay(100);
  String fname = "/image_" + String(file_number) + ".jpg";
  File file = SPIFFS.open(fname);
  boolean success = true;
  if (file) {
    long fileSize = file.size();
    Serial.println(fileSize);
    if (fileSize < (5 * 1024)) {
      success = false;
    }
    file.close();
  } else {
    success = false;
  }

  if (!success) {
    delay(500);
    SPIFFS.remove(fname);
    Serial.println(fname + " removed.");
    fname.replace("image_", "thumb_");
    SPIFFS.remove(fname);
    Serial.println(fname + " removed.");

    takePictureAndSend();
    return;
    
    if (CONNECTED) {
      SerialBT.print("filesave:false:");
      SerialBT.print("image_" + String(file_number) + ".jpg:");
       // send new line
      SerialBT.write(13);
      SerialBT.write(10);      
    }    
    
  } else {
    if (CONNECTED) {
      SerialBT.print("filesave:true:");
      SerialBT.print("image_" + String(file_number) + ".jpg:");
       // send new line
      SerialBT.write(13);
      SerialBT.write(10);      
    }    
    saveLastFileNumber();  
  }

  sensor_t * s3 = esp_camera_sensor_get();
  s3->set_framesize(s3, FRAMESIZE_HQVGA);  
    
  needDisplayRefresh = true;
  delay(500);

  filesList();
  
  if (CONNECTED) {
    tft.fillRect(0, 0, 240, 240, BLACK);
    writeText(40, 60, 1, WHITE, "Camera Remote Bluetooth");
    writeText(40, 100, 2, WHITE, "Connected.");
    writeText(40, 160, 2, WHITE, String(file_count) + " images");
    delay(100);
  } else {
    tft.fillRect(0, 0, 240, 240, BLACK);
   }  
  
  holdSendingFrame = false;
}


void loop() {

  if (CONNECTED) {


    while(SerialBT.available())
    {
      
      String received = SerialBT.readStringUntil('\n');
      Serial.println(received);
      if (received.startsWith("flash")) {
        changeFlashMode(received);
      } else if (received.startsWith("exposure")) {
        changeExposure(received);
      } else if (received.startsWith("effect")) {
        changeEffect(received);
      }
      if (!holdSendingFrame) {
        if (received.startsWith("framesize")) {
          changeFrameSize(received);
        } else if (received.startsWith("frameget")) {
          holdSendingFrame = true;
          getLastFileNumber();
          getFileNumberForImage(received);
          delay(50);
          parseData(received);
          String saveToCam = dataValues[2];
          String exposure = dataValues[3];    // ignore when take pictures
          if (saveToCam.equalsIgnoreCase("true")) {
            takePictureAndSave();  
          } else {
            takePictureAndSend();
          }
          // when space is limited, image will be sent to remote
        } else if (received.startsWith("filelist")) {
          holdSendingFrame = true;
          sendFileList();       
        } else if (received.startsWith("filecontents")) {
          holdSendingFrame = true;
          sendFileContents(received, false);
        } else if (received.startsWith("imagecontents")) {
          holdSendingFrame = true;
          sendFileContents(received, true);
        } else if (received.startsWith("filedelete")) {
          holdSendingFrame = true;
          deleteFileFromRemote(received);
        }
      } else {
        delay(100);
      }
    }

    if (!holdSendingFrame) {
      // connected, get camera frame data and send to remote android app
      esp_err_t result = camera_capture();
    }

  } else { //end of: if (CONNECTED)

    // not connected, display camera frame to display
    displayCamImage();

    processAction();
   
  }

  delay(50);
}
